import logging
import logging.config

from alembic import context
from nextlinegraphql.config import load_settings
from nextlinegraphql.hook import load_plugins
from sqlalchemy import create_engine

from nextline_rdb import models

settings = load_settings(hook=load_plugins())
url = settings.db.url

# Config in alembic.ini, only used to configure logger
config = context.config

# E.g., how to access to a config value
# script_location = config.get_main_option("script_location")

if config.config_file_name:
    # from logging_tree import printout
    # printout()
    if "nextlinegraphql" not in logging.root.manager.loggerDict:
        # Presumably, the alembic command is being executed. If programmatically
        # called, "nextlinegraphql" is in loggerDict and logging shouldn't be
        # configured here because it will override the logging configuration.
        # TODO: rearrange how configuration files are read so that this
        # conditional configuration of logging becomes cleaner or deleted.
        logging.config.fileConfig(config.config_file_name)


logger = logging.getLogger(__name__)
logger.info(f'DB URL: {url}')


target_metadata = models.Base.metadata


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = create_engine(url)

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            render_as_batch=True,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import pytest

from nextline_rdb.pagination import load_models

from .models import Entity


def test_all(session):
    Model = Entity
    models = load_models(session, Model, "id")
    expected = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    assert expected == [m.id for m in models]


params = [
    pytest.param(dict(first=0), [], id="zero"),
    pytest.param(dict(first=1), [1], id="one"),
    pytest.param(dict(first=3), [1, 2, 3], id="some"),
    pytest.param(dict(first=9), [1, 2, 3, 4, 5, 6, 7, 8, 9], id="one-fewer"),
    pytest.param(dict(first=10), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], id="exact"),
    pytest.param(dict(first=11), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], id="one-more"),
]


@pytest.mark.parametrize("kwargs, expected", params)
def test_forward(session, kwargs, expected):
    Model = Entity
    models = load_models(session, Model, "id", **kwargs)
    assert expected == [m.id for m in models]


params = [
    pytest.param(dict(after=1), [2, 3, 4, 5, 6, 7, 8, 9, 10], id="one"),
    pytest.param(dict(after=5), [6, 7, 8, 9, 10], id="middle"),
    pytest.param(dict(after=9), [10], id="one-before-last"),
    pytest.param(dict(after=10), [], id="last"),
    pytest.param(dict(after=7, first=1), [8], id="first-one"),
    pytest.param(dict(after=7, first=2), [8, 9], id="first-one-fewer"),
    pytest.param(dict(after=7, first=3), [8, 9, 10], id="first-exact"),
    pytest.param(dict(after=7, first=4), [8, 9, 10], id="first-one-more"),
    pytest.param(dict(after=10, first=1), [], id="first-one-after-last"),
]


@pytest.mark.parametrize("kwargs, expected", params)
def test_forward_with_after(session, kwargs, expected):
    Model = Entity
    models = load_models(session, Model, "id", **kwargs)
    assert expected == [m.id for m in models]


params = [
    pytest.param(dict(last=0), [], id="zero"),
    pytest.param(dict(last=1), [10], id="one"),
    pytest.param(dict(last=3), [8, 9, 10], id="some"),
    pytest.param(dict(last=9), [2, 3, 4, 5, 6, 7, 8, 9, 10], id="one-fewer"),
    pytest.param(dict(last=10), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], id="exact"),
    pytest.param(dict(last=11), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], id="one-more"),
]


@pytest.mark.parametrize("kwargs, expected", params)
def test_backward(session, kwargs, expected):
    Model = Entity
    models = load_models(session, Model, "id", **kwargs)
    assert expected == [m.id for m in models]


params = [
    pytest.param(dict(before=10), [1, 2, 3, 4, 5, 6, 7, 8, 9], id="one"),
    pytest.param(dict(before=5), [1, 2, 3, 4], id="middle"),
    pytest.param(dict(before=2), [1], id="one-before-first"),
    pytest.param(dict(before=1), [], id="first"),
    pytest.param(dict(before=4, last=1), [3], id="last-one"),
    pytest.param(dict(before=4, last=2), [2, 3], id="last-one-fewer"),
    pytest.param(dict(before=4, last=3), [1, 2, 3], id="last-exact"),
    pytest.param(dict(before=4, last=4), [1, 2, 3], id="last-one-more"),
    pytest.param(dict(before=1, last=1), [], id="last-one-before-last"),
]


@pytest.mark.parametrize("kwargs, expected", params)
def test_backward_with_before(session, kwargs, expected):
    Model = Entity
    models = load_models(session, Model, "id", **kwargs)
    assert expected == [m.id for m in models]



import pytest

from nextline_rdb.pagination import SortField, load_models

from .models import Entity

params = [
    pytest.param(
        dict(sort=[SortField("num")]),
        [7, 8, 9, 10, 4, 5, 6, 1, 2, 3],
    ),
    pytest.param(
        dict(sort=[SortField("num", True)]),
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    ),
    pytest.param(
        dict(sort=[SortField("txt")]),
        [1, 3, 4, 6, 7, 9, 2, 5, 8, 10],
    ),
    pytest.param(
        dict(sort=[SortField("txt", True)]),
        [2, 5, 8, 10, 1, 3, 4, 6, 7, 9],
    ),
    pytest.param(
        dict(sort=[SortField("num"), SortField("txt")]),
        [7, 9, 8, 10, 4, 6, 5, 1, 3, 2],
    ),
    pytest.param(
        dict(sort=[SortField("num")], after=5),
        [6, 1, 2, 3],
    ),
    pytest.param(
        dict(sort=[SortField("num")], after=5, first=3),
        [6, 1, 2],
    ),
    pytest.param(
        dict(sort=[SortField("num")], first=3),
        [7, 8, 9],
    ),
    pytest.param(
        dict(sort=[SortField("num", True)], after=5),
        [6, 7, 8, 9, 10],
    ),
    pytest.param(
        dict(sort=[SortField("num")], before=5),
        [7, 8, 9, 10, 4],
    ),
    pytest.param(
        dict(sort=[SortField("num")], before=5, last=3),
        [9, 10, 4],
    ),
    pytest.param(
        dict(sort=[SortField("num")], last=3),
        [1, 2, 3],
    ),
    pytest.param(
        dict(after=5),
        [6, 7, 8, 9, 10],
    ),
    pytest.param(
        dict(sort=[SortField("num"), SortField("txt")], after=5),
        [1, 3, 2],
    ),
]


@pytest.mark.parametrize("kwargs, expected", params)
def test_sort(session, kwargs, expected):
    Model = Entity
    id_field = "id"
    models = load_models(session, Model, id_field, **kwargs)
    assert expected == [m.id for m in models]



from pathlib import Path

import pytest


@pytest.fixture
def in_alembic_dir(monkeypatch: pytest.MonkeyPatch):
    import nextline_rdb

    path = Path(nextline_rdb.__file__).parent

    monkeypatch.chdir(path)



from __future__ import annotations

import base64
from functools import partial
from typing import TYPE_CHECKING, Callable, List, Optional, TypeVar

from nextline_rdb.pagination import load_models

from .connection import Connection, Edge, query_connection

if TYPE_CHECKING:
    from strawberry.types import Info

    from nextline_rdb import models as db_models


def encode_id(id: int) -> str:
    return base64.b64encode(f"{id}".encode()).decode()


def decode_id(cursor: str) -> int:
    return int(base64.b64decode(cursor).decode())


_T = TypeVar("_T")


def load_connection(
    info: Info,
    Model: db_models.ModelType,
    id_field: str,
    create_node_from_model: Callable[..., _T],
    *,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection[_T]:

    query_edges = partial(
        load_edges,
        Model=Model,
        id_field=id_field,
        create_node_from_model=create_node_from_model,
    )

    return query_connection(
        info,
        query_edges,
        before,
        after,
        first,
        last,
    )


def load_edges(
    info: Info,
    Model: db_models.ModelType,
    id_field: str,
    create_node_from_model: Callable[..., _T],
    *,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> List[Edge[_T]]:

    session = info.context["session"]

    models = load_models(
        session,
        Model,
        id_field,
        before=before if before is None else decode_id(before),
        after=after if after is None else decode_id(after),
        first=first,
        last=last,
    )

    nodes = [create_node_from_model(m) for m in models]

    edges = [Edge(node=n, cursor=encode_id(getattr(n, id_field))) for n in nodes]

    return edges



from logging import getLogger
from pathlib import Path
from typing import Mapping, MutableMapping, Optional, Tuple

from apluggy import asynccontextmanager
from dynaconf import Dynaconf, Validator
from nextlinegraphql.hook import spec
from sqlalchemy import func, select
from sqlalchemy.orm import Session
from starlette.applications import Starlette

from . import models
from .db import DB
from .schema import Mutation, Query, Subscription
from .write import write_db

HERE = Path(__file__).resolve().parent
DEFAULT_CONFIG_PATH = HERE / 'default.toml'
MIGRATION_CONFIG_PATH = HERE / 'migration.toml'

assert DEFAULT_CONFIG_PATH.is_file()
assert MIGRATION_CONFIG_PATH.is_file()

PRELOAD = (str(DEFAULT_CONFIG_PATH),)
SETTINGS = ()
VALIDATORS = (Validator("DB.URL", must_exist=True, is_type_of=str),)

SETTINGS_FOR_MIGRATION = (str(MIGRATION_CONFIG_PATH),)


class Plugin:
    @spec.hookimpl
    def dynaconf_preload(self) -> Optional[Tuple[str, ...]]:
        return PRELOAD

    @spec.hookimpl
    def dynaconf_settings_files(self) -> Optional[Tuple[str, ...]]:
        if Path.cwd() == HERE:
            return SETTINGS_FOR_MIGRATION
        return SETTINGS

    @spec.hookimpl
    def dynaconf_validators(self) -> Optional[Tuple[Validator, ...]]:
        return VALIDATORS

    @spec.hookimpl
    def configure(self, settings: Dynaconf):
        self._db = DB(settings.db['url'])

    @spec.hookimpl
    def initial_run_no(self):
        with self._db.session() as session:
            last_run = self._last_run(session)
            if last_run is None:
                return None
            else:
                return last_run.run_no + 1

    @spec.hookimpl
    def initial_script(self):
        with self._db.session() as session:
            last_run = self._last_run(session)
            if last_run is None:
                return None
            else:
                return last_run.script

    @spec.hookimpl
    def schema(self):
        return (Query, Mutation, Subscription)

    def _last_run(self, session: Session):
        stmt = select(models.Run, func.max(models.Run.run_no))
        if model := session.execute(stmt).scalar_one_or_none():
            return model
        else:
            logger = getLogger(__name__)
            msg = "No previous runs were found in the DB"
            logger.info(msg)
            return None

    @spec.hookimpl
    @asynccontextmanager
    async def lifespan(self, app: Starlette, context: Mapping):
        nextline = context['nextline']
        async with write_db(nextline, self._db):
            yield

    @spec.hookimpl
    def update_strawberry_context(self, context: MutableMapping) -> None:
        context['db'] = self._db



import pytest
from sqlalchemy import create_engine, select
from sqlalchemy.orm import sessionmaker

from .models import Base, Entity


@pytest.fixture
def session(db, sample):
    del sample
    with db() as y:
        yield y


def test_sample(db, sample):
    del sample
    Model = Entity
    with db() as session:
        stmt = select(Model)
        models = session.scalars(stmt)
        assert 10 == len(models.all())


@pytest.fixture
def sample(db):
    with db.begin() as session:
        num = [3, 3, 3, 2, 2, 2, 1, 1, 1, 1]
        txt = ["AA", "BB", "AA", "AA", "BB", "AA", "AA", "BB", "AA", "BB"]
        for i in range(10):
            model = Entity(num=num[i], txt=txt[i])
            session.add(model)


@pytest.fixture
def db(engine):
    Base.metadata.create_all(bind=engine)
    y = sessionmaker(autocommit=False, autoflush=False, bind=engine)
    return y


@pytest.fixture
def engine():
    url = "sqlite:///:memory:?check_same_thread=false"
    y = create_engine(url)
    return y



import datetime

from sqlalchemy import select

from nextline_rdb import DB
from nextline_rdb.models import Run


def test_one():
    db = DB()
    run_no = 1
    with db.session() as session:
        model = Run(
            run_no=run_no,
            state="running",
            started_at=datetime.datetime.utcnow(),
            ended_at=datetime.datetime.utcnow(),
            script="pass",
        )
        session.add(model)
        session.commit()
        assert model.id

    with db.session() as session:
        stmt = select(Run).filter_by(run_no=run_no)
        model = session.execute(stmt).scalar_one()



from async_asgi_testclient import TestClient
from nextlinegraphql.plugins.ctrl.test import run_statement
from nextlinegraphql.plugins.graphql.test import gql_request

from ..graphql import QUERY_HISTORY


async def test_one(client: TestClient):

    await run_statement(client)

    data = await gql_request(client, QUERY_HISTORY)
    # print(data["history"])
    runs = data["history"]["runs"]
    edges = runs["edges"]
    # assert 2 == len(runs)
    run = edges[1]["node"]
    assert 2 == run["runNo"]
    assert "finished" == run["state"]
    assert run["startedAt"]
    assert run["endedAt"]
    assert run["script"]
    assert not run["exception"]



import asyncio
from pathlib import Path
from typing import Optional, Set, cast

import pytest
from nextline import Nextline
from nextline.utils import agen_with_wait
from sqlalchemy.orm import Session

from nextline_rdb import DB
from nextline_rdb import models as db_models
from nextline_rdb import write_db


def test_one(db: DB, run_nextline, statement):
    del run_nextline
    with db.session() as session:
        session = cast(Session, session)
        runs = session.query(db_models.Run).all()  # type: ignore
        assert 2 == len(runs)
        run = runs[1]
        assert 2 == run.run_no
        assert run.started_at
        assert run.ended_at
        assert statement == run.script
        assert not run.exception

        traces = session.query(db_models.Trace).all()  # type: ignore
        assert 5 == len(traces)
        run_no = 2
        trace_no = 0
        for trace in traces:
            trace_no += 1
            assert trace_no == trace.trace_no
            assert run_no == trace.run_no
            assert trace.started_at
            assert trace.ended_at

        prompts = session.query(db_models.Prompt).all()  # type: ignore
        assert 58 == len(prompts)
        for prompt in prompts:
            assert run_no == prompt.run_no
            assert prompt.trace_no
            assert prompt.started_at
            assert prompt.line_no
            assert prompt.file_name
            assert prompt.event

        stdouts = session.query(db_models.Stdout).all()  # type: ignore
        # assert 1 == len(stdouts)
        # for stdout in stdouts:
        #     assert run_no == stdout.run_no
        #     assert stdout.text
        #     assert stdout.written_at


@pytest.fixture
def monkey_patch_syspath(monkeypatch):
    here = Path(__file__).resolve().parent
    path = here / 'example_script'
    monkeypatch.syspath_prepend(str(path))
    yield


@pytest.fixture
def statement(monkey_patch_syspath):
    del monkey_patch_syspath
    here = Path(__file__).resolve().parent
    path = here / 'example_script'
    return (path / 'script.py').read_text()


@pytest.fixture
async def run_nextline(db: DB, statement):
    nextline = Nextline(statement)
    async with write_db(nextline, db):
        async with nextline:
            task_control = asyncio.create_task(control_execution(nextline))
            await run_statement(nextline, statement)
        await task_control


@pytest.fixture
def db() -> DB:
    url = 'sqlite:///:memory:?check_same_thread=false'
    return DB(url=url)


async def run_statement(nextline: Nextline, statement: Optional[str] = None):
    await asyncio.sleep(0.01)
    await nextline.reset(statement=statement)
    await asyncio.sleep(0.01)
    async with nextline.run_session():
        pass
    await asyncio.sleep(0.01)


async def control_execution(nextline: Nextline):
    prev_ids: Set[int] = set()
    agen = agen_with_wait(nextline.subscribe_trace_ids())
    async for ids_ in agen:
        ids = set(ids_)
        new_ids, prev_ids = ids - prev_ids, ids  # type: ignore

        tasks = {asyncio.create_task(control_trace(nextline, id_)) for id_ in new_ids}
        _, pending = await agen.asend(tasks)

    await asyncio.gather(*pending)  # type: ignore


async def control_trace(nextline: Nextline, trace_no):
    file_name = ""
    async for s in nextline.subscribe_prompt_info_for(trace_no):
        if not s.open:
            continue
        if not file_name == s.file_name:
            file_name = s.file_name  # type: ignore
            assert nextline.get_source(file_name)
        command = "next"
        if s.event == "line":
            line = nextline.get_source_line(
                line_no=s.line_no,
                file_name=s.file_name,
            )
            command = find_command(line) or command
        await asyncio.sleep(0.01)
        await nextline.send_pdb_command(command, s.prompt_no, trace_no)


def find_command(line: str) -> Optional[str]:
    """The Pdb command indicated in a comment

    For example, returns "step" for the line "func()  # step"
    """
    import re

    if not (comment := extract_comment(line)):
        return None
    regex = re.compile(r"^# +(\w+) *$")
    match = regex.search(comment)
    if match:
        return match.group(1)
    return None


def extract_comment(line: str) -> Optional[str]:
    import io
    import tokenize

    comments = [
        val
        for type, val, *_ in tokenize.generate_tokens(io.StringIO(line).readline)
        if type == tokenize.COMMENT
    ]
    if comments:
        return comments[0]
    return None
